@@ -10,14 +10,12 @@ from struct   import error  as struct_error
 from numpy import dtype       as numpy_dtype
 from numpy import result_type as numpy_result_type
 
-from ..ancillaryvariables  import AncillaryVariables
 from ..coordinate          import DimensionCoordinate, AuxiliaryCoordinate
 from ..coordinatebounds    import CoordinateBounds
 from ..cellmeasure         import CellMeasure
 from ..coordinatereference import CoordinateReference
-from ..field               import Field
+from ..field               import Field, FieldList
 from ..cellmethods         import CellMethods
-from ..fieldlist           import FieldList
 from ..units               import Units
 from ..functions           import abspath, dirname, pathjoin
 
@@ -26,7 +24,7 @@ from ..data.data import Data
 from .functions import _open_netcdf_file
 from .filearray import NetCDFFileArray
 
-def read(filename, fmt=None, verbose=False):
+def read(filename, fmt=None, promote=(), verbose=False):
     ''' 
 
 Read fields from an input netCDF file on disk or from an OPeNDAP
@@ -52,6 +50,11 @@ coordinate bounds, cell measures and coordinate referencess).
         CFA-netCDF file. By default a file of any of these formats is
         read.
 
+    promote : sequence of str, optional
+
+    verbose : bool, optional
+        If True then print information to stdout.
+    
 :Returns:
 
     out : FieldList
@@ -86,6 +89,25 @@ coordinate bounds, cell measures and coordinate referencess).
         filename = name
     #--- End: if
     
+    # ----------------------------------------------------------------
+    # Parse promote
+    # ----------------------------------------------------------------
+    try:
+        iter(promote)
+    except TypeError:
+        raise ValueError(
+            "Can't read: Bad parameter value: promote=%r" % promote)
+            
+    if 'all' in promote:
+        promote = set(('reference', 'ancillary', 'dimension', 'auxiliary', 'measure'))
+    else:
+        if 'field' in promote:
+            promote = set(promote).union(('reference', 'ancillary'))
+        if 'coordinate' in promote:
+            promote = set(promote).union(('dimension', 'auxiliary'))
+
+    top_level = set()
+
     filename = abspath(filename)
 
     # Read the netCDF file 
@@ -167,13 +189,27 @@ coordinate bounds, cell measures and coordinate referencess).
     # variables. I.e. remove dimension and auxiliary coordinates,
     # their bounds and grid_mapping variables
     # ----------------------------------------------------------------
+    nc_dimensions = map(str, nc.dimensions)
+
     for ncvar in variables.copy():
 
         # Remove dimension coordinates and their bounds
-        if ncvar in map(str, nc.dimensions):
+        if ncvar in nc_dimensions:
 
             if ncvar in variables:
-                variables.discard(ncvar)
+
+                if 'dimension' in promote:
+                    # Add the dimension coordinate to the set of
+                    # top-level fields, so that it doesn't get demoted
+                    # if the auxiliary coordinate is also in a
+                    # coordinate reference.
+                    top_level.add(ncvar)
+                else:
+                    # Do not promote an auxiliary coordinate to also
+                    # appear as a top-level field
+                    variables.discard(ncvar)
+
+#                variables.discard(ncvar)
                 for attr in ('bounds', 'climatology'):
                     if attr not in attributes[ncvar]:
                         continue
@@ -202,13 +238,24 @@ locals())
             continue
         #--- End: if
 
-        # Still here? Then remove auxiliary coordinates and their
-        # bounds
+        # Still here? Then remove auxiliary coordinates (unless they
+        # have been promoted) and their bounds.
         if 'coordinates' in attributes[ncvar]:
             # Allow for (incorrect) comma separated lists
             for aux in re.split('\s+|\s*,\s*', attributes[ncvar]['coordinates']):
                 if aux in variables:
-                    variables.discard(aux)
+                    
+                    if 'auxiliary' in promote:
+                        # Add the auxiliary coordinate to the set of
+                        # top-level fields, so that it doesn't get
+                        # demoted if the auxiliary coordinate is also
+                        # in a coordinate reference.
+                        top_level.add(aux)
+                    else:
+                        # Do not promote an auxiliary coordinate to
+                        # also appear as a top-level field
+                        variables.discard(aux)
+
                     for attr in ('bounds', 'climatology'):
                         if attr not in attributes[aux]:
                             continue
@@ -235,13 +282,13 @@ locals())
                 #--- End: if
             #--- End: for
         #--- End: if
-    
+
         # Remove grid mapping variables
         if 'grid_mapping' in attributes[ncvar]:
             variables.discard(attributes[ncvar]['grid_mapping'])
 
-        # Remove cell measure variables
-        if 'cell_measures' in attributes[ncvar]:
+        # Remove cell measure variables (unless they have been promoted).
+        if 'measure' not in promote and 'cell_measures' in attributes[ncvar]:
             cell_measures = re.split('\s*(\w+):\s*',
                                      attributes[ncvar]['cell_measures'])
             for ncvar in cell_measures[2::2]:
@@ -264,13 +311,6 @@ locals())
     # {'lat': <CF Coordinate: (73)>}
     seen_in_file = {}
 
-#    # Dictionary of ?
-#    #
-#    # For example:
-#    # >>> formula_terms
-#    #
-#    formula_terms = {}
-
     # Set
     #
     # For example:
@@ -281,14 +321,14 @@ locals())
     ancillary_variables = set()
     fields_in_file      = FieldList()
 
-    for data_var in variables:
+    for data_ncvar in variables:
         # Don't turn private CFA variables into fields
-        if _is_cfa_private_variable(nc.variables[data_var], cfa):
+        if _is_cfa_private_variable(nc.variables[data_ncvar], cfa):
             continue
 
         f = _create_Field(filename,
                           nc,
-                          data_var,
+                          data_ncvar,
                           attributes,
                           seen_in_file,
                           ancillary_variables, 
@@ -304,27 +344,29 @@ locals())
     # Find which fields are being pointed to from coordinate
     # references
     # ----------------------------------------------------------------
+    if 'reference' in promote:
+        # Promote coordinate conversion fields to also appear at the
+        # top level
+        top_level.update(coordref_field_pointers)
+
     ncvar_to_field = {}
     if coordref_field_pointers:
         i = 0        
         while i < len(fields_in_file):
             f = fields_in_file[i]
             ncvar = f.ncvar            
-            if ncvar in coordref_field_pointers:
-                # Remove coordinate references from a field which is
-                # pointed to from a coordinate reference and also
-                # remove the field from the list of output fields
-                for key, coordref in f.refs().iteritems():
-                    for term in set(coordref).difference(coordref.coord_terms):
-                        if isinstance(coordref[term], dict):
-                            f.remove_item(key)
-
-                fields_in_file.pop(i)
-                # Map the pointer to the field so that we'll later be
+            if ncvar in coordref_field_pointers:   
+                # Map the pointer to the field so that later we'll be
                 # able to replace the pointer with the field
-                ncvar_to_field[ncvar] = f
-            else:
-                i += 1
+                if ncvar not in ncvar_to_field:
+                    ncvar_to_field[ncvar] = f.copy()
+
+                if ncvar not in top_level:
+                    del fields_in_file[i]
+                    i -= 1
+            #--- End: if
+
+            i += 1
     #--- End: if
 
     # ----------------------------------------------------------------
@@ -333,50 +375,72 @@ locals())
     # ----------------------------------------------------------------
     if ncvar_to_field:
         for f in fields_in_file:
-            for coordref in f.refs().values():
-                for term, value in coordref.iteritems():
+            for key, ref in f.refs().items():
+                for term, value in ref.iteritems():
                     if isinstance(value, dict):
-                        coordref[term] = ncvar_to_field[value['ncvar']].copy()
+                        # Define this coordinate conversion field
+                        try:
+                            g = ncvar_to_field[value['ncvar']].copy()
+                            # If this coordinate conversion field has a
+                            # coordinate reference which contains itself
+                            # then remove that coordinate reference.
+                            for key2, ref2 in g.refs().iteritems():
+                                for value2 in ref2.itervalues():
+                                    if isinstance(value2, dict) and value2['ncvar'] == g.ncvar:
+                                        g.remove_item(key2)
+                            #--- End: for
+                            ref[term] = g
+                        except KeyError:
+                            # Something went wrong. Most likely an
+                            # inappropriate coordinate reference
+                            # belonging to a promotes dimension or
+                            # auxiliary corodinate. Just get rid of
+                            # the coordinate reference.
+                            f.remove_item(key)
+                            break
     #--- End: if
         
     # ----------------------------------------------------------------
-    # For each field that has ancillary variables, replace its
-    # built-in list of netCDF variable names with an
-    # AncillaryVariables object.
+    # For each field that has ancillary variables, replace its list of
+    # netCDF variable names with a FieldList object.
     # ----------------------------------------------------------------
+    if 'ancillary' in promote:
+        # Promote ancillary fields to also appear at the top level
+        top_level.update(ancillary_variables)
+
     if ancillary_variables:
         ncvar_to_field = {}
-        index = 0
+        i = 0
         while ancillary_variables:
             try:
-                f = fields_in_file[index]
+                f = fields_in_file[i]
             except IndexError:
                 # No more fields
                 break
             
-            # Still here?
             ncvar = f.ncvar            
             if ncvar in ancillary_variables:
-                # ----------------------------------------------------
                 # This field is being used as an ancillary variable in
                 # another field
-                # ----------------------------------------------------
                 ancillary_variables.discard(ncvar)
                 ncvar_to_field[ncvar] = f
-                fields_in_file.pop(index)
-            else:
-                index += 1
+                if ncvar not in top_level:                    
+                    del fields_in_file[i]
+                    i -= 1 
+            #--- End: if
+
+            i += 1
         #--- End: while
         
         for f in fields_in_file:
             if not hasattr(f, 'ancillary_variables'):
                 continue
 
-            av = AncillaryVariables()
-            for ncvar in f.ancillary_variables:
-                av.append(ncvar_to_field[ncvar].copy())
+            av = [ncvar_to_field[ncvar].copy() 
+                  for ncvar in f.ancillary_variables]
 
-            f.ancillary_variables = av
+#            f.ancillary_variables = AncillaryVariables(av)
+            f.ancillary_variables = FieldList(av)
         #--- End: for
     #--- End: if
 
@@ -385,7 +449,7 @@ locals())
 
 def _create_Field(filename,
                   nc,
-                  data_var,
+                  data_ncvar,
                   attributes,
                   seen_in_file,
                   ancillary_variables, 
@@ -405,16 +469,14 @@ Create a field for a given netCDF variable.
     nc : netCDF4.Dataset
         The entire netCDF file in a `netCDF4.Dataset` instance.
 
-    data_var : str
-        The netCDF name of the variable to be turned into a field.
+    data_ncvar : str
+        The name of the netCDF variable to be turned into a field.
 
     attributes : dict
         Dictionary of the data variable's netCDF attributes.
 
     seen_in_file : dict
 
-#    formula_terms : dict
-
     ancillary_variables : set
 
     global_attributes : dict
@@ -429,14 +491,12 @@ Create a field for a given netCDF variable.
         The new field.
 
 '''
-    properties = attributes[data_var]
+    properties = attributes[data_ncvar]
 
     # Add global attributes to the data variable's properties, unless
     # the data variables already has a property with the same name.
     for attr, value in global_attributes.iteritems():
-#        if attr not in attributes[data_var]:
         if attr not in properties:
-#            attributes[data_var][attr] = value
             properties[attr] = value
 
     # Take cell_methods out of the data variable's properties since it
@@ -474,8 +534,8 @@ cell_methods)
 
     # Change numpy arrays to tuples for selected attributes
     for attr in ('valid_range',):
-#        if attr in attributes[data_var]:
-#            attributes[data_var][attr] = tuple(attributes[data_var][attr])
+#        if attr in attributes[data_ncvar]:
+#            attributes[data_ncvar][attr] = tuple(attributes[data_ncvar][attr])
         if attr in properties:
             properties[attr] = tuple(properties[attr])
 
@@ -487,7 +547,7 @@ cell_methods)
 
     f = Field(properties=properties, copy=False)
 
-    f.ncvar = data_var
+    f.ncvar = data_ncvar
     f.file  = filename
     f.Units = f_Units
 
@@ -506,7 +566,7 @@ cell_methods)
     # ----------------------------------------------------------------
     # Add axes and non-scalar dimension coordinates to the field
     # ----------------------------------------------------------------
-    field_ncdimensions = _ncdimensions(nc.variables[data_var], cfa)
+    field_ncdimensions = _ncdimensions(nc.variables[data_ncvar], cfa)
 
     for ncdim in field_ncdimensions:
         if ncdim in nc.variables:
@@ -528,7 +588,7 @@ cell_methods)
             # just create a dimension with the correct size.
             dim = f.domain.insert_axis(size=len(nc.dimensions[ncdim]))
         #--- End: if
-
+        
         # Update data dimension name and set dimension size
         f.domain.nc_dimensions[dim] = ncdim
         f.domain._axes['data'].append(dim)
@@ -536,7 +596,7 @@ cell_methods)
         ncdim_to_dim[ncdim] = dim
     #--- End: for
     
-    f.Data = _set_Data(nc, nc.variables[data_var], f, f,
+    f.Data = _set_Data(nc, nc.variables[data_ncvar], f, f,
                        unpacked_dtype=unpacked_dtype, cfa=cfa)
 
     # ----------------------------------------------------------------
@@ -616,99 +676,9 @@ cell_methods)
 
         _create_formula_terms_ref(f, key, coord, formula_terms,
                                   attributes, ncvar_to_key,
-                                  coordref_field_pointers)
+                                  coordref_field_pointers, seen_in_file)
     #--- End: for
 
-#        # Add the equation terms and references to their values to to
-#        # new auxiliary coordinate's coordinate reference.
-#        kwargs = {}
-#        ft = re.split('\s+|\s*:\s+', form_terms)
-#        for term, ncvar in zip(ft[0::2], ft[1::2]):
-#            if ncvar in ncvar_to_key:
-#                # CASE 1: This variable is a coordinate of the field,
-#                #         so we point to it from the coordinate reference.
-#                value = ncvar_to_key[ncvar]
-#
-#            elif ncvar not in attributes:
-#                # CASE 2: The value does not exist as a netCDF
-#                #         variable in this file
-#                value = None
-#
-#            else:
-#                # CASE 3: This variable is not a coordinate of the
-#                #         field so it goes in to the coordinate reference as an
-#                #         independent field
-#
-##            elif 'bounds' not in attributes[ncvar]:
-##                # CASE 3: This variable is not a coordinate of the
-##                #         field and it has no bounds, so it goes in to
-##                #         the coordinate reference as an independent field.
-#                value = {'ncvar': ncvar}
-#                coordref_field_pointers.add(ncvar)
-#                
-##            else:
-##                # CASE 4: This variable is not a coordinate of the
-##                #         field and yet it has bounds
-##                aux_dims = []
-##                var_ncdimensions = _ncdimensions(nc.variables[ncvar], cfa)
-##                for ncdim in var_ncdimensions:
-##                    if ncdim not in ncdim_to_dim:
-##                        aux_dims = None
-##                        break
-##
-##                    aux_dims.append(ncdim_to_dim[ncdim])
-##                #--- End: for
-##
-##                if aux_dims is None:
-##                    # CASE 4a: This variable's dimensions are not a
-##                    #          subset of the field's dimensions, so it
-##                    #          can't be a coordinate. Therefore it
-##                    #          goes into the coordinate reference as an
-##                    #          independent field with the bounds
-##                    #          deleted.
-##                    value = {'ncvar': ncvar}
-##
-##                else:
-##                    # CASE 4b: This variable's dimensions are a subset
-##                    #          of the field's dimensions, so we may
-##                    #          allow it to be a coordinate. Therefore
-##                    #          add it to the field as an auxiliary
-##                    #          coordinate and point to it from the
-##                    #          coordinate reference.
-##                    coord = _create_Coordinate(nc, ncvar, attributes, f, cfa=cfa,
-##                                               dimension=False)
-##                    if coord.hasprop('formula_terms'):
-##                        coord.delprop('formula_terms')
-##
-##                    aux = f.domain.insert_aux(coord, axes=aux_dims, key=aux, 
-##                                              copy=False)
-##
-##                    ncvar_to_key[ncvar] = aux
-##
-##                    # Add this variable name and its bounds name to
-##                    # the set of variables which are not fields. Note
-##                    # that if it turns out that this variable is also
-##                    # being used as a field in another coordinate reference then
-##                    # its name will be removed from this set.
-##                    not_fields.update((ncvar, attributes[ncvar]['bounds']))
-##                    
-##                    value = aux
-#            #--- End: if
-#                    
-#            kwargs[term] = value
-#        #--- End: for 
-#
-#        name = coord.getprop('standard_name', None)
-#        coordref = CoordinateReference(name=name, coords=(key,), crtype='formula_terms',
-#                                        **kwargs)
-#
-#        f.domain.insert_coordref(coordref, copy=False)
-#
-#        formula_terms[coord_ncvar] = coordref
-#
-##        coord.delprop('formula_terms')
-#    #--- End: for
-
     # ----------------------------------------------------------------
     # Add grid mapping coordinate references
     # ----------------------------------------------------------------
@@ -716,21 +686,6 @@ cell_methods)
     if grid_mapping is not None:
         _create_grid_mapping_ref(f, grid_mapping, attributes, ncvar_to_key)
 
-#        for ncvar in grid_mapping.split():
-#            if ncvar not in attributes:
-#                continue
-
-#            name = attributes[ncvar].pop('grid_mapping_name', None)
-#
-#            coordref = CoordinateReference(name, crtype='grid_mapping',
-#                                            **attributes[ncvar])
-#            coordref.ncvar = ncvar
-#
-#            coordref_id = f.domain.insert_coordref(coordref, copy=False)
-#        #--- End: for
-#
-#        f.delprop('grid_mapping')
-
     # ----------------------------------------------------------------
     # Add cell measures to the field
     # ----------------------------------------------------------------
@@ -776,7 +731,7 @@ cell_methods)
  
     # ----------------------------------------------------------------
     # Parse an ancillary_variables string to a list of netCDF variable
-    # names, which will get converted to an AncillaryVariables object
+    # names, which will get converted to an FieldList object
     # later. Add these netCDF variable names to the set of all
     # ancillary data variables in the file.
     # ----------------------------------------------------------------
@@ -785,7 +740,7 @@ cell_methods)
         ancillary_variables.update(f.ancillary_variables)
     #--- End: if
     
-    f.setcyclic()
+    f.autocyclic()
 
     # Return the finished field
     return f
@@ -1042,7 +997,7 @@ def _create_grid_mapping_ref(f, grid_mapping, attributes, ncvar_to_key):
 
 def _create_formula_terms_ref(f, key, coord, formula_terms,
                               attributes, ncvar_to_key,
-                              coordref_field_pointers):
+                              coordref_field_pointers, seen_in_file):
     '''
 
 :Parameters:
@@ -1054,6 +1009,7 @@ def _create_formula_terms_ref(f, key, coord, formula_terms,
     coord : cf.Coordinate
 
     formula_terms : str
+        The formula_terms attribute value from the netCDF file.
 
     attributes : dict
 
@@ -1065,30 +1021,38 @@ def _create_formula_terms_ref(f, key, coord, formula_terms,
 
     out : cf.CoordinateReference
 
-'''        
+'''
+    standard_name = coord.getprop('standard_name', None)
+
     # Add the equation terms and references to their values to to
     # new auxiliary coordinate's coordinate reference.
     kwargs      = {}
     coord_terms = []
 
     ft = re.split('\s+|\s*:\s+', formula_terms)
+    ncvars = ft[1::2]
 
-    for term, ncvar in zip(ft[0::2], ft[1::2]):
+    for term, ncvar in zip(ft[0::2], ncvars):
         if ncvar in ncvar_to_key:
-            # CASE 1: This variable is a coordinate of the field, so
-            #         we point to it from the coordinate reference.
+            # CASE 1: The term's value is a coordinate of the field,
+            #         so we point to it from the coordinate reference.
             value = ncvar_to_key[ncvar]
             coord_terms.append(term)
 
+##            # Make sure that coordinate has an identity
+#            c = seen_in_file[ncvar]
+#            if standard_name is not None and not hasattr(c, 'standard_name'):
+#                c.id = standard_name + '_formula_term_' + term
+
         elif ncvar not in attributes:
-            # CASE 2: The value does not exist as a netCDF variable in
-            #         this file
+            # CASE 2: The term's value does not exist as a netCDF
+            #         variable in this file
             value = None
             
         else:
-            # CASE 3: This variable is not a coordinate of the field
-            #         so it goes in to the coordinate reference as an
-            #         independent field
+            # CASE 3: The term's value is not a coordinate of the
+            #         field so it goes into the coordinate reference
+            #         as an independent field
             value = {'ncvar': ncvar}
             coordref_field_pointers.add(ncvar)
         #--- End: if
@@ -1096,7 +1060,7 @@ def _create_formula_terms_ref(f, key, coord, formula_terms,
         kwargs[term] = value
     #--- End: for 
 
-    coordref = CoordinateReference(coord.getprop('standard_name', None),
+    coordref = CoordinateReference(standard_name,
                                    crtype='formula_terms',
                                    coords=(key,),
                                    coord_terms=coord_terms,
@@ -1143,7 +1107,7 @@ Set the Data attribute of a variable.
             cfa_data = json_loads(variable.getprop('cfa_array'))
         except ValueError as error:
             raise ValueError(
-                "Error during JSON-decoding of 'cfa_array' property: %s" %
+                "Error during JSON-decoding of netCDF attribute 'cfa_array': %s" %
                 error)
 
         cfa_data['file']       = f.file
