@@ -88,15 +88,15 @@ Any amount of any combination of CF-netCDF, CFA-netCDF and Met Office
     umversion : int or float, optional
         Met Office (UK) PP files and Met Office (UK) fields files
         only, the Unified Model (UM) version to be used when decoding
-        the header. Valid versions are, for example, ``402``,
-        ``'606.3'`` and ``'1001'``.  The default version is
-        ``405``. The version is ignored if it can be inferred from the
-        headers (which is generally true for PP files created at
-        version 5.3 and later and for all fields files). Note that
-        the PP header can not encode tertiary version elements (such
-        as the ``3`` in ``606.3``), so it may be necessary to provide
-        a UM version in such cases.
-
+        the PP header. Valid versions are, for example, ``402`` or
+        ``4.2`` (v4.2), ``606.3`` (v6.6.3) and ``1001`` or ``10.1``
+        (v10.1). The default version is ``4.5`` (v4.5). The version is
+        ignored if it can be inferred from the PP headers, which will
+        generally be the case for files created at v5.3 and
+        later. Note that the PP header can not encode tertiary version
+        elements (such as the ``3`` in ``606.3``), so it may be
+        necessary to provide a UM version in such cases.
+    
         Ignored for any other type of input files.
 
     ignore_read_error : bool, optional
@@ -213,8 +213,7 @@ Any amount of any combination of CF-netCDF, CFA-netCDF and Met Office
     
     # Parse umversion
     if umversion is not None:
-        umversion = str(umversion)
-        umversion = umversion.replace('.', '0', 1)
+        umversion = float(str(umversion).replace('.', '0', 1))
 
     # ----------------------------------------------------------------
     # Parse top_level
